/****************************************************************
* Daniel Greco
* driver.cpp contains the main function, as well as all the graphics 
* code that will be needed to run the program.
*
*****************************************************************/
#include "Input.h"
#include "Points.h"
#include "HelperFunctions.h"
#include "Lighting.h"


GLuint vbo, colorbuffer, normalbuffer, shineBuffer, uvbuffer;
GLuint mv, proj, v, ts, rot;
Points *points;
Input *input;
Light *lighting;
mat4 projection, rotate, translate, view;
GLfloat fov, aspect;
GLfloat zNear, zFar;

GLuint horizAngle, vertAngle;


//generateBuffers is used to initialize all the buffers.  Another init procedure
void generateBuffers()
{
	glGenBuffers(1, &vbo);
	glGenBuffers(1, &colorbuffer);
	glGenBuffers(1, &normalbuffer);
	glGenBuffers(1, &shineBuffer);
	glGenBuffers(1, &uvbuffer);

}

//display will render all the vertices to the screen.
void display()
{
	
	//clear the buffers for color writing
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	
	/**********
	Note once working, move these enables into a seperate function!
	They don't need to be called for every display, since nothing is updating!
	*************/

	//first attrib array is the points of the triangles
    glEnableVertexAttribArray(0);
    glBindBuffer(GL_ARRAY_BUFFER, vbo);
    glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 0, 0);

	//second attrib array are the colors.
	glEnableVertexAttribArray(1);
	glBindBuffer(GL_ARRAY_BUFFER, colorbuffer);
	//Note: Not sure why i'm casting 0 to a void pointer, consider removing
	glVertexAttribPointer(1, 3, GL_FLOAT, GL_FALSE, 0, (void *)0);

	//third attrib array are the normals
	glEnableVertexAttribArray(2);
	glBindBuffer(GL_ARRAY_BUFFER, normalbuffer);
	glVertexAttribPointer(2, 3, GL_FLOAT, GL_FALSE, 0, (void *)0);

	//fourth are the shines for the cubes
	glEnableVertexAttribArray(3);
	glBindBuffer(GL_ARRAY_BUFFER, shineBuffer);
	glVertexAttribPointer(3, 1, GL_FLOAT, GL_FALSE, 0, (void *)0);



	

	//create view vector components
	vec4 eye = vec4(10,10, 0, 0);
	vec4 at = vec4(0, 0 , 0, 1);
	vec4 up = vec4(0,1,0,0);
	

	//create the view matrix
	view = LookAt(eye, at, up);

	//pass the model matricies to the gpu
	glUniformMatrix4fv(ts, 1, GL_TRUE, translate);
	glUniformMatrix4fv(rot, 1, GL_TRUE, rotate);

	glUniformMatrix4fv(v, 1, GL_TRUE, view); 
	//I'm currently using perspective, but frustum could be used also
	//Note that frustum does do some cool things that perspective doesn't
	//projection = Frustum(1, -1, -1, 1, zNear, zFar);
	projection = Perspective(fov,aspect, zNear, zFar);
	glUniformMatrix4fv(proj, 1, GL_TRUE, projection);


	//updateLighting every frame.  However know that lighting will only ever be updated when the user does something to cause the change.
	lighting->updateLighting();

	//The program is split up into two different parts
    //draw a weird shape by sharing vertices, saves space
	glDrawArrays(GL_TRIANGLES, 0, points->getVecSize());

	
	//disable the arrays and get ready for the next draw
    glDisableVertexAttribArray(0);
	glDisableVertexAttribArray(1);
	glDisableVertexAttribArray(2);
	glDisableVertexAttribArray(3);

    glutSwapBuffers();
}

void reshape(int width, int height)
{
	//change the viewport to be the new width and height
	glViewport(0, 0, width, height);
	//change the aspect ratio now.
	aspect = GLfloat(width)/height;
}

/********
Mouse, keyboard and special keys now are wrapped using helper methods
*********/
void	mouse(const int button, const int state, const int x, const int y)
{
	input->mouse(button, state, x, y);
}
void	keyboard(const unsigned char key, const int x, const int y)
{
	input->keyboard(key, x, y);
}

void specialkeys(const int key, const int x, const int y)
{
	input ->specialkeys(key, x, y);
}

//initcallbacks will pass all the functions that we need to glut
void initCallbacks()
{
	glutDisplayFunc( display );
    glutKeyboardFunc( keyboard );
	glutMouseFunc( mouse);
	glutReshapeFunc( reshape);
	glutSpecialFunc(specialkeys);
}

void initMatricies()
{
	//note angle will complain but we don't want the matricies to be anything initially so set them to identity
	projection = identity();
	rotate = identity();
	translate = identity();
	view = identity();
	

	fov = 45.0f;
	zNear = .5;
	//bumped up zFar so none of our squares get clipped out of frame
	zFar = 50;
}


//Init will set up the shaders as well as get the window ready for display
//It will also set up all the callback functions.
void init ()
{

    // Load shaders and use the resulting shader program
    GLuint program = InitShader( "vshader.glsl", "fshader.glsl" );
    glUseProgram( program );


	
	makeFarTooManyCubes(150);
	//createLighting(program);
	lighting->init(program);

	//if textures work use this otherwise ignore
	/*createCheckerCube(0.0, WHITE);
	createTexture(program);
	*/
    // Initialize the vertex position attribute from the vertex shader
    GLuint loc = glGetAttribLocation( program, "vPosition" );
	mv = glGetUniformLocation(program, "modelview");
	proj = glGetUniformLocation(program, "projection");
	v = glGetUniformLocation(program, "view");
	rot =  glGetUniformLocation(program, "rotate");
	ts =  glGetUniformLocation(program, "translate");
	
	initMatricies();

    glClearColor( 0, 0, 0, 0 ); // black background
	//saves some computation
	//this line commented out caused bugs, as the cube would lose sides, due to the culling.
	//glEnable(GL_CULL_FACE);
	glEnable(GL_DEPTH_TEST);
	initCallbacks();
}




bool initGlut(int argc, char **argv)
{
	//initGlut
	glutInit( &argc, argv);
	glutInitDisplayMode( GLUT_RGBA | GLUT_DEPTH | GLUT_DOUBLE);  //render in rgba with double buffering
	glutInitWindowSize(SCREEN_WIDTH, SCREEN_HEIGHT);
	//position the window 10x10 away from the top left corner. This was done to just mess around with glut
	glutInitWindowPosition(10,10);
	glutInitContextVersion( 3, 1 ); //nothing drawn from above 3.1 
	//use freeglut to get a glut profile from version 4 of opengl
    glutInitContextProfile( GLUT_CORE_PROFILE );
	//finally create the window with the name assignment 2 in the title
	glutCreateWindow( "Lighting and More!!(Hopefully)" );

	//as of right now i have no way to know if something fails so return true always
	//if things can fail look into that
	return true;
}

//this will be run once we are done with the program
void cleanup()
{
	delete points;
	delete input;
	delete lighting;
}


int main(int argc, char **argv)
{
	points = new Points();
	input = new Input();
	lighting = new Light();

	//make sure points was set up correctly, otherwise the program won't work
	if(!points->init())
	{
		std::cerr << "Could not initialize points class" << std::endl;
		return -2;
	}

	//old code that's still hanging in there. It's not needed but doesn't cause problems
	input->setPart(InputNS::PART_1);

	//set up all the glut functions
	initGlut(argc, argv);

	//set up glew
	GLenum status =  glewInit();

	//make sure glew is set up properly
	if(status != GLEW_OK)
	{
		std::cerr << "GLEW not initialized correctly error is " << glewGetErrorString(status) <<  std::endl;
		return -1;
	}

	
    init();


	//run the loop that will never end
    glutMainLoop();

	//hopefully we get here although i don't think we do
	cleanup();

	return 0;
}