/************************************************************************
Daniel Greco
Helper functions for the project includes:
creating shapes (Cubes)
handling mouse clicks
************************************************************************/
#include <algorithm> //for max
#include <cmath>
#include "HelperFunctions.h"
#include "Points.h"


//buffer objects
extern GLuint colorbuffer, vbo, normalbuffer, shineBuffer, uvbuffer;
extern Points * points;


//translates mouse clicks to open gl coordinates
void screenToOpenGL(int x, int y)
{
	//get the viewpoint coordinates
	GLint viewport[4];
	glGetIntegerv(GL_VIEWPORT, viewport);

	//get the model view matrix
	GLdouble modelview[16];
	glGetDoublev(GL_MODELVIEW_MATRIX, modelview);

	//get the projection matrix
	GLdouble projection[16];
	glGetDoublev(GL_PROJECTION_MATRIX, projection);

	//convert the mouse clicks to windows coordinates
	GLfloat winX, winY, winZ;
	winX = (float)x;
	winY = (float)y;
	//move y into the viewport if it isn't there
	winY = (float)viewport[3] - winY;

	//read in a bunch of pixels, starting from the mouse coordinate and going to the right and bottom of the screen
	//it will store all of these in winz
	glReadPixels(x, y, 1, 1, GL_DEPTH_COMPONENT, GL_FLOAT, &winZ);

	GLdouble posX, posY, posZ;

	//map windows coordinates(0-width, 0-height) to object coordinates(-1-1, -1-1).  The result will be stored in posx, posy, posz
	gluUnProject( winX, winY, winZ, modelview, projection, viewport, &posX, &posY, &posZ);

	//if you'd like to send the points over immedietely feel free to uncomment them. But don't.
	//add the list and generate a random color vector 
	//points->addToList(vec3(posX, posY, posZ), randomColor());
}

//create a random color for a vertex
vec3 randomColor()
{
	GLfloat red, green, blue;
	red = (float)(rand() % 100) / 100;
	green = (float)(rand() % 100) / 100;
	blue = (float)(rand() % 100) / 100;

	return vec3(red, green, blue);
}




//quad will allow you to make one side of a cube.
void quad(vec3 uLeft, vec3 uRight, vec3 lLeft, vec3 lRight, vec3 color)
{
	//subtract two sides of the triangle
	vec3 u = lRight - lLeft;
	vec3 v = uRight - lRight;

	//compute the normal of the cross product to get the normal of this quad
	vec3 normal = normalize(cross(u, v));

	//save all the points.  These points are two triangles. Note it was done this way,
	//because we are using GlTriangles to draw, so to avoid worrying about shapes
	points->addToList(lLeft, color, normal);
	points->addToList(lRight, color, normal);
	points->addToList(uRight, color, normal);
	points->addToList(lLeft, color, normal);
	points->addToList(uRight,color , normal);
	points->addToList(uLeft, color, normal);

}

//the first createcube attempt. saved for backwards compatibility 
void createCube()
{
	//if you want one simple cube, we'll give you one centered at the origin, with a red color
	createCube(0.0, RED);
}

//this updated version of createcube will simply offset the x position, and again pass the color along. Another backwards compatible function
void createCube(GLfloat offset, vec3 color)
{
	createCube(offset, 0.0, 0.0, color);	
}

//this function will now allow you to shift by x y and z, allowing you to get some cubes everywhere
void createCube(GLfloat xoffset, GLfloat yoffset, GLfloat zoffset, vec3 color)
{
	//create 8 vectors, each one shifted however the offset is
	vec3 pt0 = vec3( -0.5 + xoffset, -0.5 + yoffset,  0.5 + zoffset);
    vec3 pt1 = vec3( -0.5 + xoffset,  0.5 + yoffset,  0.5 + zoffset);
    vec3 pt2 = vec3(  0.5 + xoffset,  0.5 + yoffset,  0.5 + zoffset);
    vec3 pt3 = vec3(  0.5 + xoffset, -0.5 + yoffset,  0.5 + zoffset);
    vec3 pt4 = vec3( -0.5 + xoffset, -0.5 + yoffset, -0.5 + zoffset);
    vec3 pt5 = vec3( -0.5 + xoffset,  0.5 + yoffset , -0.5 + zoffset);
    vec3 pt6 = vec3(  0.5 + xoffset,  0.5 + yoffset, -0.5 + zoffset);
    vec3 pt7 = vec3(  0.5 + xoffset, -0.5 + yoffset, -0.5 + zoffset);


	//make each side of the cube.
	quad(pt1, pt2, pt0, pt3, color);
	quad(pt2, pt6, pt3, pt7, color);
	quad(pt4, pt7, pt0, pt3, color);
	quad(pt5, pt6, pt1, pt2, color);
	quad(pt5, pt6, pt4, pt7, color);
	quad(pt1, pt5, pt0, pt4, color);



}

//the final version of create cube.  It's identical to the one seen above, but now it also stores a shine factor as well
void createCube(GLfloat xoffset, GLfloat yoffset, GLfloat zoffset, vec3 color, GLfloat shine)
{
	createCube(xoffset, yoffset, zoffset, color);
	points->addShine(shine);
}

//this helper function sends all the points over to the GPU.  This is done because we are create a bunch of cubes at once, and 
//rather then sending after every cube creation, we send it over when we're good and ready
void sendToGPU()
{
	//update the arrays
	points->purgeQueue();

	//pass over the vertex points
	glGenBuffers(1, &vbo);
	glBindBuffer(GL_ARRAY_BUFFER, vbo);
	glBufferData(GL_ARRAY_BUFFER, sizeof(points->getVectors()->getVecArray()[0]) * points->getVecSize(), points->getVectors()->getVecArray(), GL_STATIC_DRAW);
	
	//send over the colors
	glGenBuffers(1, &colorbuffer);
	glBindBuffer(GL_ARRAY_BUFFER, colorbuffer);
	glBufferData(GL_ARRAY_BUFFER, sizeof(points->getVectors()->getColorArray()[0]) * points->getVecSize(), points->getVectors()->getColorArray(), GL_STATIC_DRAW);

	//how about some normals
	glGenBuffers(1, &normalbuffer);
	glBindBuffer(GL_ARRAY_BUFFER, normalbuffer);
	glBufferData(GL_ARRAY_BUFFER, sizeof(points->getVectors()->getNormalArray()[0]) * points->getVecSize(), points->getVectors()->getNormalArray(), GL_STATIC_DRAW);

	//send over the shine coords
	glGenBuffers(1, &shineBuffer);
	glBindBuffer(GL_ARRAY_BUFFER, shineBuffer);
	glBufferData(GL_ARRAY_BUFFER, sizeof(points->getVectors()->getShineArray()[0]) * points->getVecSize(), points->getVectors()->getShineArray(), GL_STATIC_DRAW);
}


//this function just creates a whole bunch of cubes (150 typically) of random positions, and shine.  Note this function
//will also take care of sending over the cubes to the GPU
void makeFarTooManyCubes(int numCubes)
{
	//if the user gives bad input yell
	if(numCubes <= 0)
	{
		std::cerr<< "Bad cube variable" << std::endl;
		return;
	}

	//make three in an easy to see place just to make sure 
	createCube();
	createCube(-2.5, BLUE);
	createCube(2.5, GREEN);

	//make all those cubes with those random variables.  Note that 15 is how much we want to allow the cubes to move.  If you 
	//get too big it's hard to find the cubes, so stick to a lower number, but not too low or they all overlap.
	for(int i = 0; i < numCubes; ++i)
	{
		int rX = rand() % 15;
		int rY = rand() % 15;
		int rZ = rand() % 15;
		//shine gets crazy the higher you go so just cap it at 200, as we get high white colors when we get those high end variables
		int shine = rand() % 200;
		//make the cube!
		createCube(rX, rY, rZ, randomColor(), shine);
	}
	sendToGPU();

}

/*********************
NOTHING USED BELOW IGNORE!
Note quadTexture and createCheckeredCube would be used if texturing was working.  It's nearly identical but uses uv coordinates in addition
to all the other things that are computed

*********************/


void quadTexture(vec3 uLeft, vec3 uRight, vec3 lLeft, vec3 lRight, vec3 color)
{
		//subtract 
	vec3 u = lRight - lLeft;
	vec3 v = uRight - lRight;

	vec3 normal = normalize(cross(u, v));



	points->addToList(lLeft, color, normal, vec2(0.0, 0.0));
	points->addToList(lRight, color, normal, vec2(0.0, 1.0));
	points->addToList(uRight, color, normal, vec2(1.0, 1.0));
	points->addToList(lLeft, color, normal, vec2(0.0,0.0));
	points->addToList(uRight,color , normal, vec2(1.0, 1.0));
	points->addToList(uLeft, color, normal, vec2(0.0, 1.0));

}

void createCheckerCube(GLfloat offset, vec3 color)
{
	vec3 pt0 = vec3( -0.5 + offset, -0.5,  0.5);
    vec3 pt1 = vec3( -0.5 + offset,  0.5 ,  0.5);
    vec3 pt2 = vec3(  0.5 + offset,  0.5 ,  0.5);
    vec3 pt3 = vec3(  0.5 + offset, -0.5 ,  0.5);
    vec3 pt4 = vec3( -0.5 + offset, -0.5 , -0.5);
    vec3 pt5 = vec3( -0.5 + offset,  0.5  , -0.5);
    vec3 pt6 = vec3(  0.5 + offset,  0.5 , -0.5);
    vec3 pt7 = vec3(  0.5 + offset, -0.5, -0.5);

	quadTexture(pt1, pt2, pt0, pt3, color);
	quadTexture(pt2, pt6, pt3, pt7, color);
	quadTexture(pt4, pt7, pt0, pt3, color);
	quadTexture(pt5, pt6, pt1, pt2, color);
	quadTexture(pt5, pt6, pt4, pt7, color);
	quadTexture(pt1, pt5, pt0, pt4, color);

	points->purgeQueue();
	

	//note since our array is dynamically created, we can't just pass the size of the array, we have to get the size
	//of one item in the array and multiply it by the size of the entire array.
	//send over the vertices
	glGenBuffers(1, &vbo);
	glBindBuffer(GL_ARRAY_BUFFER, vbo);
	glBufferData(GL_ARRAY_BUFFER, sizeof(points->getVectors()->getVecArray()[0]) * points->getVecSize(), points->getVectors()->getVecArray(), GL_STATIC_DRAW);
	
	//send over the colors
	glGenBuffers(1, &colorbuffer);
	glBindBuffer(GL_ARRAY_BUFFER, colorbuffer);
	glBufferData(GL_ARRAY_BUFFER, sizeof(points->getVectors()->getColorArray()[0]) * points->getVecSize(), points->getVectors()->getColorArray(), GL_STATIC_DRAW);

	//normals!
	glGenBuffers(1, &normalbuffer);
	glBindBuffer(GL_ARRAY_BUFFER, normalbuffer);
	glBufferData(GL_ARRAY_BUFFER, sizeof(points->getVectors()->getNormalArray()[0]) * points->getVecSize(), points->getVectors()->getNormalArray(), GL_STATIC_DRAW);

	//uvs if you comment out this line you will crash my program. Don't do it please!!
	/*glGenBuffers(1, &uvbuffer);
	glBindBuffer(GL_ARRAY_BUFFER, uvbuffer);
	glBufferData(GL_ARRAY_BUFFER, sizeof(points->getVectors()->getUVSArray()[0]) * points->getVecSize(), points->getVectors()->getUVSArray(), GL_STATIC_DRAW);*/

}

//Compute color will be used when creating the torus. It will smoothly blend from red to green to blue.
vec3 computeColor(int alpha)
{
	GLfloat red, green, blue;
	//make sure we handle when it's at 0-120 and 240 - 360
	if(alpha <= 120)
		red = std::max<float>(0, 1 - fabs(-((double)alpha / 120)));
	else
		red = std::max<float>(0, 1 - fabs(-(((double)alpha - 360)/ 120)));
	green = std::max<float>(0, 1 - fabs((((double)alpha - 120) / 120)));
	blue = std::max<float>(0, 1 - fabs((((double)alpha - 240) / 120)));
	
	return vec3(red, green, blue);
}


//not used in the project
void createTorus()
{
	
	vec3 temp;
	double alpha, beta;
	
	
	for(int i = 0; i <= 36; ++i)
	{
		//move 10 degrees
		alpha = i * 10* TORAD;
		
		for(int j = 0; j <= 36; ++j)
		{
			//move 10 degrees
			beta =  j * 10*  TORAD;
			//create a vector of the current circle.
			temp = vec3(
				cos(alpha) * (CIRCLE_RADIUS + SECTION_RADIUS*cos(beta)), 
				SECTION_RADIUS * sin(beta),
				sin(alpha) * (CIRCLE_RADIUS + SECTION_RADIUS*cos(beta)));
			points->addToList(temp,computeColor(i * 10));
			
			//create a second vector of the adjacent circle, so we can draw the triangle nicely. 
			temp = vec3(
				cos(alpha + 10 * TORAD) * (CIRCLE_RADIUS + SECTION_RADIUS*cos(beta)), 
				SECTION_RADIUS * sin(beta),
				sin(alpha + 10 * TORAD) * (CIRCLE_RADIUS + SECTION_RADIUS*cos(beta)));
			points->addToList(temp,computeColor(i * 10 + 10));
		}
	}
	
	//add all the points to the draw arrays
	points->purgeQueue();
	glGenBuffers(1, &vbo);
	glBindBuffer(GL_ARRAY_BUFFER, vbo);
	
	//note since our array is dynamically created, we can't just pass the size of the array, we have to get the size
	//of one item in the array and multiply it by the size of the entire array.
	glBufferData(GL_ARRAY_BUFFER, sizeof(points->getVectors()->getVecArray()[0]) * points->getVecSize(), points->getVectors()->getVecArray(), GL_STATIC_DRAW);
	glGenBuffers(1, &colorbuffer);
	glBindBuffer(GL_ARRAY_BUFFER, colorbuffer);
	glBufferData(GL_ARRAY_BUFFER, sizeof(points->getVectors()->getColorArray()[0]) * points->getVecSize(), points->getVectors()->getColorArray(), GL_STATIC_DRAW);


}